local pairs = pairs;
local ipairs = ipairs;

-- empties tbl and assigns the value true to each key passed as part of ...
local function fillTable( tbl, ... )
    for key in pairs(tbl) do
        tbl[key] = nil;
    end
    for i = 1, select("#", ...), 1 do
        local key = select(i, ...);
        key = tonumber(key) or key;
        tbl[key] = true;
    end
end

-- same as fillTable() except that each key is also stored in
-- the array portion of the table in order
local function doubleFillTable( tbl, ... )
    fillTable(tbl, ...);
    for i = 1, select("#", ...), 1 do
        tbl[i] = select(i, ...);
    end
end

--working tables
local tokenTable = {};
local sortingTable = {};
local groupingTable = {};
local tempTable = {};

local function GetGroupHeaderType(self)
    local type, start, stop;

    local nRaid = GetNumRaidMembers();
    local nParty = GetNumPartyMembers();
    if ( nRaid > 0 and self:GetAttribute("showRaid") ) then
        type = "RAID";
    elseif ( (nRaid > 0 or nParty > 0) and self:GetAttribute("showParty") ) then
        type = "PARTY";
    elseif ( self:GetAttribute("showSolo") ) then
        type = "SOLO";
    end
    if ( type ) then
        if ( type == "RAID" ) then
            start = 1;
            stop = nRaid;
        else
            if ( type == "SOLO" or self:GetAttribute("showPlayer") ) then
                start = 0;
            else
                start = 1;
            end
            stop = nParty;
        end
    end
    return type, start, stop;
end

local function GetGroupRosterInfo(type, index)
    local _, unit, name, subgroup, className, role;
    if ( type == "RAID" ) then
        unit = "raid"..index;
        name, _, subgroup, _, _, className, _, _, _, role = GetRaidRosterInfo(index);
    else
        if ( index > 0 ) then
            unit = "party"..index;
        else
            unit = "player";
        end
        if ( UnitExists(unit) ) then
            name = UnitName(unit);
            _, className = UnitClass(unit);
            if ( GetPartyAssignment("MAINTANK", unit) ) then
                role = "MAINTANK";
            elseif ( GetPartyAssignment("MAINASSIST", unit) ) then
                role = "MAINASSIST";
            end
        end
        subgroup = 1;
    end
    return unit, name, subgroup, className, role;
end

function SecureGroupHeader_UpdateCopy(self)
    local nameList = self:GetAttribute("nameList");
    local groupFilter = self:GetAttribute("groupFilter");
    local sortMethod = self:GetAttribute("sortMethod");
    local groupBy = self:GetAttribute("groupBy");

    for key in pairs(sortingTable) do
        sortingTable[key] = nil;
    end

    -- See if this header should be shown
    local type, start, stop = GetGroupHeaderType(self);
    if ( not type ) then
        configureChildren(self);
        return;
    end

    if ( not groupFilter and not nameList ) then
        groupFilter = "1,2,3,4,5,6,7,8";
    end

    if ( groupFilter ) then
        -- filtering by a list of group numbers and/or classes
        fillTable(tokenTable, strsplit(",", groupFilter));
        local strictFiltering = self:GetAttribute("strictFiltering"); -- non-strict by default
        for i = start, stop, 1 do
            local unit, name, subgroup, className, role = GetGroupRosterInfo(type, i);
            if ( name and
                ((not strictFiltering) and
                 (tokenTable[subgroup] or tokenTable[className] or (role and tokenTable[role])) -- non-strict filtering
             ) or
                (tokenTable[subgroup] and tokenTable[className]) -- strict filtering
            ) then
                tinsert(sortingTable, name);
                sortingTable[name] = unit;
                if ( groupBy == "GROUP" ) then
                    groupingTable[name] = subgroup;

                elseif ( groupBy == "CLASS" ) then
                    groupingTable[name] = className;

                elseif ( groupBy == "ROLE" ) then
                    groupingTable[name] = role;

                end
            end
        end

        if ( groupBy ) then
            local groupingOrder = self:GetAttribute("groupingOrder");
            doubleFillTable(tokenTable, strsplit(",", groupingOrder));
            for k in pairs(tempTable) do
                tempTable[k] = nil;
            end
            for _, grouping in ipairs(tokenTable) do
                grouping = tonumber(grouping) or grouping;
                for k in ipairs(groupingTable) do
                    groupingTable[k] = nil;
                end
                for index, name in ipairs(sortingTable) do
                    if ( groupingTable[name] == grouping ) then
                        tinsert(groupingTable, name);
                        tempTable[name] = true;
                    end
                end
                if ( sortMethod == "NAME" ) then -- sort by ID by default
                    table.sort(groupingTable);
                end
                for _, name in ipairs(groupingTable) do
                    tinsert(tempTable, name);
                end
            end
            -- handle units whose group didn't appear in groupingOrder
            for k in ipairs(groupingTable) do
                groupingTable[k] = nil;
            end
            for index, name in ipairs(sortingTable) do
                if not ( tempTable[name] ) then
                    tinsert(groupingTable, name);
                end
            end
            if ( sortMethod == "NAME" ) then -- sort by ID by default
                table.sort(groupingTable);
            end
            for _, name in ipairs(groupingTable) do
                tinsert(tempTable, name);
            end

            --copy the names back to sortingTable
            for index, name in ipairs(tempTable) do
                sortingTable[index] = name;
            end

        elseif ( sortMethod == "NAME" ) then -- sort by ID by default
            table.sort(sortingTable);

        end

    else
        -- filtering via a list of names
        doubleFillTable(sortingTable, strsplit(",", nameList));
        for i = start, stop, 1 do
            local unit, name = GetGroupRosterInfo(type, i);
            if ( sortingTable[name] ) then
                sortingTable[name] = unit;
            end
        end
        for i = #sortingTable, 1, -1 do
            local name = sortingTable[i];
            if ( sortingTable[name] == true ) then
                tremove(sortingTable, i);
            end
        end
        if ( sortMethod == "NAME" ) then
            table.sort(sortingTable);
        end
    end

    return sortingTable;
end